/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <fcall.h>
#include <ip.h>
#include <9p.h>
#include <mp.h>
#include <libsec.h>

#define VERSION "ATA over Ethernet file server for Plan 9"

enum
{
	Blocksize = 512,

	Cread = 	0x24,
	Cwrite = 	0x34,
};



uvlong mlba;
ushort major;
uchar minor;
uvlong size;
char firmware[9];
char model[41];
char serial[21];

char *dest;
char *ethern;
char *user;
uint tag;
int debug;
int netfd;
int forcemulti;
int blocks;
ulong time0;

void
usage(void)
{
	fprint(2, "usage: %s [-Ddf] [-a major] [-b nblocks] [-e ethern] [-i minor] [-m mtpt]"
			" [-s srvname] dest\n", argv0);
	exits("usage");
}

enum
{
	Qroot,
	Qctl,
	Qdata,
};

#define PATH(type, n)		((type)|((n)<<8))
#define TYPE(path)			((int)(path) & 0xFF)
#define NUM(path)			((uint)(path)>>8)

typedef struct aoehdr aoehdr;
struct aoehdr {
	uchar edst[6];
	uchar esrc[6];
	uchar type[2];
	char flags;
	char error;
	char major[2];
	char minor;
	char cmd;
	char tag[4];
	char data[1477];
};

typedef struct aoeata aoeata;
struct aoeata {
	char aflags;
	char error;
	char sectcount;
	char cmd;
	char lba[6];
	char reserved[2];
	char data[1024];
};

typedef struct aoeqry aoeqry;
struct aoeqry {
	char bcount[2];
	char version[2];
	char reserved;
	char flags;
	char slen[2];
	char string[1025];
};

void
aoestrings(char *s)
{
	int i;
	char a;

	for(i = 0; i < strlen(s) - strlen(s) % 2; i += 2) {
		a = s[i];
		s[i] = s[i + 1];
		s[i + 1] = a;
	}
}

uvlong
getlba(char *data)
{
	uvlong lba;

	memmove(&lba, data, sizeof(lba));

	return lba;
}

void
setlba(char *data, uvlong lba)
{

	memmove(data, &lba, sizeof(lba));
}

int
recvaoe(aoehdr *pkt)
{

	memset(pkt, 0, sizeof(aoehdr));
	if(read(netfd, pkt, sizeof(aoehdr)) < 0)
		sysfatal("read: %r");

	if(pkt->flags & 0x84) {
		fprint(2, "AoE error: ");
		switch(pkt->error) {
		case 0x01:
			fprint(2, "Unrecognized command\n");
			break;
		case 0x02:
			fprint(2, "Bad argument parameter\n");
			break;
		case 0x03:
			fprint(2, "Device unavailable\n");
			break;
		case 0x04:
			fprint(2, "Config string present\n");
			break;
		case 0x05:
			fprint(2, "Unsupported version\n");
			break;
		default:
			fprint(2, "Unknown error code\n");
			break;
		};
		return 1;
	}

	return 0;
}	

uvlong
getaoesize(void)
{
	aoehdr pkt;
	aoeata *ata;
	uvlong ret;
	uchar ttag[4];

	memset(&pkt, 0, sizeof(aoehdr));
	parseether(pkt.edst, dest);
	hnputs(pkt.type, 0x88A2);
	hnputs(pkt.major, major);
	pkt.minor = minor;
	pkt.flags = 0x10;
	hnputl(ttag, ++tag);
	memmove(pkt.tag, ttag, sizeof(pkt.tag));
	ata = (aoeata *)pkt.data;
	ata->aflags = 0x40;
	ata->sectcount = 0x01;
	ata->cmd = 0xEC;

	if(write(netfd, &pkt, 64) < 0)
		sysfatal("write: %r");

	for(;;) {
		if(recvaoe(&pkt))
			return 0;

		if(memcmp(pkt.tag, ttag, sizeof(pkt.tag)) != 0)
			continue;
		if(!(pkt.flags & 0x08) || pkt.cmd != 0x00)
			continue;

		mlba = getlba(ata->data + 200);
		ret = mlba * Blocksize;
		memmove(firmware, ata->data + 46, 8);
		aoestrings(firmware);
		memmove(model, ata->data + 54, 40);
		aoestrings(model);
		memmove(serial, ata->data + 20, 20);
		aoestrings(serial);
		break;
	}

	return ret;
}

int
discaoe(void)
{
	aoehdr pkt;
	aoeqry *qry;
	ushort len, major, firmware;

	memset(&pkt, 0, sizeof(aoehdr));
	memset(pkt.edst, 0xFF, sizeof(pkt.edst));
	hnputs(pkt.type, 0x88A2);
	hnputs(pkt.major, 0xFFFF);
	pkt.minor = 0xFF;
	pkt.cmd = 0x01;
	pkt.flags = 0x10;
	qry = (aoeqry *)pkt.data;
	hnputl(qry->slen, strlen(VERSION));
	memmove(qry->string, VERSION, strlen(VERSION));

	if(write(netfd, &pkt, 32 + strlen(VERSION)) < 0)
		sysfatal("write: %r");

	for(;;) {
		if(recvaoe(&pkt))
			continue;

		if(pkt.flags & 0x08 && pkt.cmd == 0x01) {
			len = nhgets(qry->slen);
			qry->string[len] = '\0';
			major = nhgets(pkt.major);
			firmware = nhgets(qry->version);
			fprint(2, "device: %E(%s) with major %d, minor %d, ver %d, firmware %d\n",
					pkt.esrc, qry->string, major, pkt.minor & 0xFF,
					(qry->flags >> 0x04) & 0x0F, firmware);
		}
	}
}

int
dialaoe(void)
{
	int tempfd;

	tempfd = dial(netmkaddr("0x88A2", ethern, nil), 0, 0, 0);
	if(tempfd < 0)
		sysfatal("dial: %r");

	return tempfd;
}

int
doblock(uvlong lba, uchar *block, uint nblocks, uint cmd)
{
	aoehdr pkt;
	aoeata *ata;
	char ttag[4];

	memset(&pkt, 0, sizeof(aoehdr));
	parseether(pkt.edst, dest);
	hnputs(pkt.type, 0x88A2);
	hnputs(pkt.major, major);
	pkt.minor = minor;
	pkt.flags = 0x10;
	hnputl(ttag, ++tag);
	memmove(pkt.tag, ttag, sizeof(pkt.tag));
	ata = (aoeata *)pkt.data;
	ata->aflags = 0x40;
	ata->sectcount = nblocks & 0xFF;
	ata->cmd = cmd & 0xFF;
	setlba(ata->lba, lba);
	if(cmd == Cwrite)
		memmove(ata->data, block, Blocksize * nblocks);

	if(write(netfd, &pkt, 36 + ((cmd == Cwrite) ? Blocksize * nblocks : 28)) < 0)
		sysfatal("write: %r");

	for(;;) {
		if(recvaoe(&pkt))
			return 0;

		if(memcmp(pkt.tag, ttag, sizeof(pkt.tag)) != 0)
			continue;
		if(!(pkt.flags & 0x08) || pkt.cmd != 0x00)
			continue;

		if(ata->cmd & 0x01)
			return 0;

		if(cmd == Cread)
			memmove(block, ata->data, Blocksize * nblocks);

		break;
	}

	return 1;
}

void
doaoereq(Req *r, uint cmd)
{
	uvlong n, i, b, e, bc;
	uchar *block;

	if(r->ifcall.offset > size) {
		respond(r, "block device is smaller");
		return;
	}

	if(r->ifcall.count + r->ifcall.offset >= size)
		r->ifcall.count = size - r->ifcall.offset;
	if(r->ifcall.count == 0) {
		respond(r, nil);
		return;
	}

	block = emalloc9p(Blocksize * blocks);

	i = r->ifcall.offset % (uvlong)Blocksize;
	n = (r->ifcall.offset - i) / (uvlong)Blocksize;
	if(i != 0) {
		i = Blocksize - i;
		doblock(n, block, 1, Cread);

		if(cmd == Cread)
			memmove(r->ofcall.data, block + (Blocksize - i), i);
		else {
			memmove(block, r->ifcall.data, i);
			doblock(n, block, 1, Cwrite);
		}
		n++;
	}

	e = (r->ifcall.count - i) % Blocksize;
	b = n + ((r->ifcall.count - i - e) / (uvlong)Blocksize);
	for(bc = 1; n < b; n += bc, i += Blocksize * bc) {
		if(b - n >= 1) {
			if(b - n > blocks)
				bc = blocks;
			else
				bc = b - n;
		}
		if(cmd == Cwrite)
			memmove(block, r->ifcall.data + i, Blocksize * bc);
		doblock(n, block, bc, cmd);
		if(cmd == Cread)
			memmove(r->ofcall.data + i, block, Blocksize * bc);
	}

	if(e != 0) {
		doblock(n, block, 1, Cread);

		if(cmd == Cread)
			memmove(r->ofcall.data + i, block, e);
		else {
			memmove(block, r->ifcall.data + i, e);
			doblock(n, block, 1, Cwrite);
		}
	}
	r->ofcall.count = r->ifcall.count;

	free(block);

	respond(r, nil);
}

typedef struct Tab Tab;
struct Tab
{
	char *name;
	ulong mode;
};

Tab tab[] =
{
	".",		DMDIR|0555,
	"ctl",		0666,
	"data",		0777,
};

static void
fillstat(Dir *d, uvlong path)
{
	Tab *t;

	memset(d, 0, sizeof(*d));
	d->uid = estrdup9p(user);
	d->gid = estrdup9p(user);
	d->qid.path = path;
	d->atime = d->mtime = time0;
	t = &tab[TYPE(path)];
	d->name = estrdup9p(t->name);
	if(!strncmp(d->name, "data", 4))
		d->length = size;
	else
		d->length = 0;
	d->qid.type = t->mode>>24;
	d->mode = t->mode;
}

static void
fsattach(Req *r)
{
	if(r->ifcall.aname && r->ifcall.aname[0]){
		respond(r, "invalid attach specifier");
		return;
	}
	r->fid->qid.path = PATH(Qroot, 0);
	r->fid->qid.type = QTDIR;
	r->fid->qid.vers = 0;
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}

static void
fsstat(Req *r)
{
	fillstat(&r->d, r->fid->qid.path);
	respond(r, nil);
}

static int
rootgen(int i, Dir *d, void*)
{
	i += Qroot + 1;

	if(i <= Qdata){
		fillstat(d, i);
		return 0;
	}
	return -1;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	int i;
	ulong path;

	path = fid->qid.path;
	if(!(fid->qid.type & QTDIR))
		return "walk in non-directory";

	if(strcmp(name, "..") == 0){
		switch(TYPE(path)){
		case Qroot:
			return nil;
		default:
			return "bug in fswalk1";
		}
	}

	i = TYPE(path) + 1;
	while(i < nelem(tab)){
		if(strcmp(name, tab[i].name) == 0){
			qid->path = PATH(i, NUM(path));
			qid->type = tab[i].mode>>24;
			return nil;
		}
		if(tab[i].mode & DMDIR)
			break;
		i++;
	}
	return "directory entry not found";
}

static void
fsopen(Req *r)
{
	respond(r, nil);
}

static void
fsread(Req *r)
{
	char e[ERRMAX], buf[8192];
	ulong path;
	int n;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qroot:
		dirread9p(r, rootgen, nil);
		respond(r, nil);
		break;
	case Qctl:
		n = snprint(buf, sizeof(buf), "inquiry %s %s\n"
					"geometry %lld %d %lld %lld %lld\n", model, serial, size,
					Blocksize, (mlba & 0x00FF0000) >> 16,
					(mlba & 0x0000FF00) >> 8, mlba & 0x000000FF);
		readbuf(r, buf, n);
		respond(r, nil);
		break;
	case Qdata:
		doaoereq(r, Cread);
		break;
	default:
		snprint(e, sizeof(e), "bug in fsread path=%lux", path);
		respond(r, e);
		break;
	}
}

static void
fswrite(Req *r)
{
	char e[ERRMAX];
	ulong path;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qctl:
		respond(r, nil);
		break;
	case Qdata:
		doaoereq(r, Cwrite);
		break;
	default:
		snprint(e, sizeof(e), "bug in fswrite path=%lux", path);
		respond(r, e);
		break;
	}
}

Srv fs = 
{
.attach=	fsattach,
.walk1=		fswalk1,
.open=		fsopen,
.read=		fsread,
.write=		fswrite,
.stat=		fsstat,
};

void
threadmain(int argc, char **argv)
{
	char *mtpt, *srvname;

	mtpt = "/n/aoe";
	srvname = nil;
	major = -1;
	minor = -1;
	ethern = "ether0";
	dest = "ffffffffffff";
	blocks = 2;

	fmtinstall('E', eipfmt);

	ARGBEGIN {
	case 'D':
		chatty9p++;
		break;
	case 'd':
		debug++;
		break;
	case 'f':
		forcemulti++;
		break;
	case 's':
		srvname = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 'e':
		ethern = EARGF(usage());
		break;
	case 'a':
		major = (ushort)atoi(EARGF(usage())) & 0xFFFF;
		break;
	case 'i':
		minor = (uchar)atoi(EARGF(usage())) & 0xFF;
		break;
	case 'b':
		blocks = atoi(EARGF(usage()));
		break;
	default:
		usage();
	} ARGEND;

	netfd = dialaoe();
	if(argc < 1) {
		discaoe();
		threadexits(0);
	}

	time0 = time(0);
	if(!forcemulti)
		dest = argv[0];
	user = getuser();
	size = getaoesize();

	threadpostmountsrv(&fs, srvname, mtpt, MBEFORE);
	threadexits(0);
}
